----------------------------------------------------------------------------------
-- Company: 
-- Engineer: 
-- 
-- Create Date:    12:39:40 06/03/2015 
-- Design Name: 
-- Module Name:    MPAread - Behavioral 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
use IEEE.std_logic_misc.all;

-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
--use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx primitives in this code.
-- June-3-2015
-- SPI mode fixed as 1
-- 6 MPAs are daisy chained, length for each MPA increased by one because of the dummy bit added at the MSB
-- row2 pixel ordering reversed in saving to counter_rbuf
library UNISIM;
use UNISIM.VComponents.all;
Library UNIMACRO;
use UNIMACRO.vcomponents.all;

entity MPAread is
    Port ( reset : in  STD_LOGIC;
					 sysclk : in  STD_LOGIC;
					 set_read_clk : in  STD_LOGIC;
					 read_start : in  STD_LOGIC;
					 read_all : in  STD_LOGIC;
           read_addr : in  STD_LOGIC_VECTOR(1 downto 0);
					 read_clk_p : out  STD_LOGIC;
           read_clk_n : out  STD_LOGIC;
					 read_din_p : in  STD_LOGIC;
           read_din_n : in  STD_LOGIC;
					 read_dout_p : out  STD_LOGIC;
           read_dout_n : out  STD_LOGIC;
           read_done : out  STD_LOGIC;
           mem_p : in  STD_LOGIC_VECTOR(5 downto 0);
           mem_n : in  STD_LOGIC_VECTOR(5 downto 0);
           nMPA : in  STD_LOGIC_VECTOR(2 downto 0);
           ipb_clk : in  STD_LOGIC;
					 ipb_write : in std_logic;
					 ipb_strobe : in std_logic;
					 ipb_addr : in std_logic_vector(15 downto 0);
					 ipb_wdata : in std_logic_vector(31 downto 0);
           ipb_rdata : out  STD_LOGIC_VECTOR (31 downto 0));
end MPAread;

architecture Behavioral of MPAread is
constant N : integer := 3;
signal MODE : std_logic_vector(1 downto 0) := "01";
constant zero : std_logic_vector(31 downto 0) := (others =>'0');
type array6x1 is array (0 to 5) of std_logic_vector(0 downto 0);
type array6x32 is array (0 to 5) of std_logic_vector(31 downto 0);
signal mem_rbuf_DI : array6x1 := (others => (others =>'0'));
signal mem_rbuf_DO : array6x32 := (others => (others =>'0'));
signal read_clk : std_logic := '0';
signal read_din : std_logic := '0';
signal read_dout : std_logic := '0';
signal mem_in : std_logic_vector(5 downto 0) := (others =>'0');
signal ipb_buf_NUM : std_logic_vector(1 downto 0) := (others =>'0');
signal buf_NUM : std_logic_vector(1 downto 0) := (others =>'0');
signal nMPAsub1 : std_logic_vector(2 downto 0) := (others =>'0');
signal DIV : std_logic_vector(N downto 0) := (others =>'0');
signal cnt : std_logic_vector(N downto 0) := (others =>'0');
signal read_clk_o : std_logic := '0';
signal start : std_logic := '0';
signal ipb_start : std_logic := '0';
signal ipb_read_all : std_logic := '0';
signal GO : std_logic := '0';
signal GO_q : std_logic := '0';
signal send : std_logic := '0';
signal rec : std_logic := '0';
signal en_rec : std_logic := '0';
signal en_counter_rec : std_logic := '0';
signal en_send : std_logic := '0';
signal toggle : std_logic := '0';
signal toggle_q : std_logic := '0';
signal strobe : std_logic_vector(2 downto 0) := (others =>'0');
signal wbuf_we : std_logic := '0';
signal row2 : std_logic := '0';
signal counter_pixel : std_logic := '0';
signal wbuf_addra : std_logic_vector(9 downto 0) := (others =>'0');
signal wbuf_ra : std_logic_vector(14 downto 0) := (others =>'0');
signal wbuf_DOA : std_logic_vector(31 downto 0) := (others =>'0');
signal wbuf_DOB : std_logic_vector(31 downto 0) := (others =>'0');
signal counter_rbuf_DO : std_logic_vector(31 downto 0) := (others =>'0');
signal counter_rbuf_DI : std_logic_vector(0 downto 0) := (others =>'0');
signal counter_rbuf_wa : std_logic_vector(14 downto 0) := (others =>'0');
signal mem_rbuf_wa : std_logic_vector(14 downto 0) := (others =>'0');
signal rbuf_wren : std_logic := '0';
signal counter_rbuf_we : std_logic_vector(0 downto 0) := (others =>'0');
signal mem_rbuf_we : std_logic_vector(0 downto 0) := (others =>'0');
attribute IOB : string;
attribute IOB of read_clk, read_dout, counter_rbuf_DI, mem_rbuf_DI : signal is "true";
begin
g_mem_in : for i in 0 to 5 generate
	i_mem_in_buf: IBUFDS generic map(DIFF_TERM => TRUE,IOSTANDARD => "LVDS_25") port map(i => mem_p(i), ib => mem_n(i), o => mem_in(i));
end generate;
i_read_din_buf: IBUFDS generic map(DIFF_TERM => TRUE,IOSTANDARD => "LVDS_25") port map(i => read_din_p, ib => read_din_n, o => read_din);
i_read_clk_obuf : OBUFDS generic map (IOSTANDARD => "LVDS_25") port map (O => read_clk_p, OB => read_clk_n, I => read_clk);
i_read_dout_obuf : OBUFDS generic map (IOSTANDARD => "LVDS_25") port map (O => read_dout_p, OB => read_dout_n, I => read_dout);
nMPAsub1 <= "000" when nMPA(2 downto 1) = "00" else nMPA - 1;
process(ipb_clk)
begin
	if(ipb_clk'event and ipb_clk = '1')then
		if(rec = '1')then
			counter_rbuf_DI(0) <= read_din;
			for i in 0 to 5 loop
				mem_rbuf_DI(i)(0) <= mem_in(i);
			end loop;
		end if;
		if(send = '1')then
			read_dout <= wbuf_DOB(conv_integer(wbuf_ra(4 downto 0)));
		end if;
		read_clk <= read_clk_o;
		if(read_start = '1')then
			mem_rbuf_we(0) <= read_all;
			buf_NUM <= read_addr;
		elsif(ipb_start = '1')then
			mem_rbuf_we(0) <= ipb_read_all;
			buf_NUM <= ipb_buf_NUM;
		end if;
		if(start = '1')then
			if(mem_rbuf_we(0) = '1')then
				wbuf_ra <= "010000111110111";
			else
				wbuf_ra <= "00" & nMPAsub1 & "1100011111";
			end if;
		elsif(send = '1')then
			if(or_reduce(wbuf_ra(9 downto 0)) = '0')then
				wbuf_ra(9 downto 0) <= "1100100000";
				wbuf_ra(13 downto 10) <= wbuf_ra(13 downto 10) - 1;
			else
				wbuf_ra(9 downto 0) <= wbuf_ra(9 downto 0) - 1;
			end if;
		end if;
		if(start = '1')then
			mem_rbuf_wa <= buf_NUM & "1101100000000";
		elsif(rbuf_wren = '1')then
			mem_rbuf_wa <= mem_rbuf_wa - 1;
		end if;
		if(start = '1')then
			counter_rbuf_wa <= buf_NUM & nMPAsub1 & "1100100000";
			row2 <= '0';
		elsif(rbuf_wren = '1')then
			if(or_reduce(counter_rbuf_wa(9 downto 0)) = '1')then
				counter_rbuf_wa(3 downto 0) <= counter_rbuf_wa(3 downto 0) - 1;
			end if;
			if(counter_rbuf_wa(3 downto 0) = x"0")then
				if(or_reduce(counter_rbuf_wa(9 downto 4)) = '0')then
					counter_rbuf_wa(12 downto 10) <= counter_rbuf_wa(12 downto 10) - 1;
					counter_rbuf_wa(9 downto 4) <= "110010";
				else
					counter_rbuf_wa(9 downto 4) <= counter_rbuf_wa(9 downto 4) - 1;
				end if;
--				elsif(row2 = '0')then
--					if(counter_rbuf_wa(9 downto 4) = "100010")then
--						counter_rbuf_wa(9 downto 4) <= "010010";
--						row2 <= '1';
--					else
--						counter_rbuf_wa(9 downto 4) <= counter_rbuf_wa(9 downto 4) - 1;
--					end if;
--				elsif(counter_rbuf_wa(9 downto 4) = "100001")then
--					counter_rbuf_wa(9 downto 4) <= "010001";
--					row2 <= '0';
--				else
--					counter_rbuf_wa(9 downto 4) <= counter_rbuf_wa(9 downto 4) + 1;
--				end if;
			end if;
		end if;
	end if;
end process;
process(ipb_clk,reset,set_read_clk)
begin
	if(reset = '1')then
		read_clk_o <= '0';
	elsif(set_read_clk = '1')then
		read_clk_o <= '1';
	elsif(ipb_clk'event and ipb_clk = '1')then
		if(start = '1')then
			read_clk_o <= '1';
		elsif(or_reduce(cnt(N downto 1)) = '0')then
			if(GO = '1')then
				read_clk_o <= not read_clk_o;
			end if;
		end if;
	end if;
end process;
process(ipb_clk,reset)
begin
	if(reset = '1')then
		DIV   <= (others =>'0');
		cnt   <= (others =>'0');
		ipb_start    <= '0';
		ipb_read_all    <= '0';
		ipb_buf_NUM   <= (others =>'0');
		start    <= '0';
		GO    <= '0';
		GO_q    <= '0';
		send     <= '0';
		rec      <= '0';
		en_rec <= '0';
		en_send <= '0';
		rbuf_wren <= '0';
		counter_pixel <= '1';
		read_done <= '0';
	elsif(ipb_clk'event and ipb_clk = '1')then
		if(ipb_addr = x"9c00" and ipb_write = '1' and ipb_strobe = '1')then
			ipb_start <= '1';
			ipb_read_all <= ipb_wdata(0);
			ipb_buf_NUM <= ipb_wdata(2 downto 1);
		else
			ipb_start <= '0';
		end if;
		if(ipb_addr = x"9c02" and ipb_write = '1' and ipb_strobe = '1')then
			DIV <= ipb_wdata(N downto 0);
		end if;
		if(start = '1')then
			GO <= '1';
		elsif(or_reduce(cnt(N downto 1)) = '0' and en_send = '0' and (MODE(0) = '1' or en_rec = '0'))then
			GO <= '0';
		end if;
		start <= read_start or ipb_start;
		GO_q <= GO;
		if(start = '1')then
			cnt <= DIV;
		elsif(GO = '1' or or_reduce(cnt(N downto 1)) = '1')then
			if(or_reduce(cnt(N downto 1)) = '0')then
				cnt <= DIV;
			else 
				cnt <= cnt - 1;
			end if;
		end if;
		if(start = '1')then
			counter_rbuf_we <= "1";
		elsif(or_reduce(counter_rbuf_wa(12 downto 0)) = '0' and rbuf_wren = '1')then
			counter_rbuf_we <= "0";
		end if;
		if(start = '1')then
			en_rec <= '1';
		elsif(or_reduce(cnt(N downto 1)) = '0' and en_send = '0' and xor_reduce(MODE & read_clk_o) = '0')then
			en_rec <= '0';
		end if;
		if(start = '1')then
			en_send <= '1';
		elsif(or_reduce(cnt(N downto 1)) = '0' and or_reduce(wbuf_ra(13 downto 0)) = '0' and xor_reduce(MODE & read_clk_o) = '1')then
			en_send <= '0';
		end if;
		if(or_reduce(cnt(N downto 1)) = '0' and en_rec = '1' and xor_reduce(MODE & read_clk_o) = '0')then
			rec <= '1';
		else
			rec <= '0';
		end if;
		if(start = '1' and MODE(0) = '0')then
			send <= '1';
		elsif(or_reduce(cnt(N downto 1)) = '0' and en_send = '1' and xor_reduce(MODE & read_clk_o) = '1')then
			send <= '1';
		else
			send <= '0';
		end if;
		rbuf_wren <= rec;
		read_done <= not GO and GO_q;
		if(start = '1')then
			counter_pixel <= '1';
		elsif(send = '1' and wbuf_ra(4 downto 0) = "00000")then
			if(wbuf_ra(9 downto 5) = "00001")then
				counter_pixel <= '0';
			else
				counter_pixel <= '1';
			end if;
		end if;
	end if;
end process;
process(ipb_clk,reset)
begin
	if(reset = '1')then
		toggle <= '0';
	elsif(ipb_clk'event and ipb_clk = '1')then
		toggle <= not toggle;
	end if;
end process;
process(sysclk,reset)
begin
	if(reset = '1')then
		toggle_q <= '0';
		strobe <= (others => '0');
		wbuf_we <= '0';
	elsif(sysclk'event and sysclk = '1')then
		toggle_q <= toggle;
		strobe <= strobe(1 downto 0) & (toggle_q xor toggle);
		if(ipb_addr(15 downto 4) = x"9c1" and ipb_write = '1' and ipb_strobe = '1' and strobe(2) = '1')then
			wbuf_we <= '1';
		else
			wbuf_we <= '0';
		end if;		
	end if;
end process;
g_wbuf : for i in 0 to 31 generate
   i_wbuf : RAM32X1D
   port map (
      DPO => wbuf_DOB(i),     -- Read-only 1-bit data output
      SPO => wbuf_DOA(i),     -- R/W 1-bit data output
      A0 => ipb_addr(0),       -- R/W address[0] input bit
      A1 => ipb_addr(1),       -- R/W address[1] input bit
      A2 => ipb_addr(2),       -- R/W address[2] input bit
      A3 => ipb_addr(3),       -- R/W address[3] input bit
      A4 => '0',       -- R/W address[4] input bit
      D => ipb_wdata(i),         -- Write 1-bit data input
      DPRA0 => wbuf_ra(10), -- Read-only address[0] input bit
      DPRA1 => wbuf_ra(11), -- Read-only address[1] input bit
      DPRA2 => wbuf_ra(12), -- Read-only address[2] input bit
      DPRA3 => counter_pixel, -- Read-only address[3] input bit
      DPRA4 => '0', -- Read-only address[4] input bit
      WCLK => ipb_clk,   -- Write clock input
      WE => wbuf_we        -- Write enable input
   );
end generate;
i_counter_rbuf : BRAM_SDP_MACRO
   generic map (
      BRAM_SIZE => "36Kb", -- Target BRAM, "18Kb" or "36Kb" 
      DEVICE => "VIRTEX6", -- Target device: "VIRTEX5", "VIRTEX6", "SPARTAN6" 
      WRITE_WIDTH => 1,    -- Valid values are 1-72 (37-72 only valid when BRAM_SIZE="36Kb")
      READ_WIDTH => 32)     -- Valid values are 1-72 (37-72 only valid when BRAM_SIZE="36Kb")
    port map (
      DO => counter_rbuf_DO,         -- Output read data port, width defined by READ_WIDTH parameter
      DI => counter_rbuf_DI,         -- Input write data port, width defined by WRITE_WIDTH parameter
      RDADDR => ipb_addr(9 downto 0), -- Input read address, width defined by read port depth
      RDCLK => sysclk,   -- 1-bit input read clock
      RDEN => '1',     -- 1-bit input read port enable
      REGCE => '0',   -- 1-bit input read output register enable
      RST => '0',       -- 1-bit input reset 
      WE => counter_rbuf_we,         -- Input write enable, width defined by write port depth
      WRADDR => counter_rbuf_wa, -- Input write address, width defined by write port depth
      WRCLK => ipb_clk,   -- 1-bit input write clock
      WREN => rbuf_wren      -- 1-bit input write port enable
   );
g_mem_rbuf : for i in 0 to 5 generate
	i_mem_rbuf : BRAM_SDP_MACRO
   generic map (
      BRAM_SIZE => "36Kb", -- Target BRAM, "18Kb" or "36Kb" 
      DEVICE => "VIRTEX6", -- Target device: "VIRTEX5", "VIRTEX6", "SPARTAN6" 
      WRITE_WIDTH => 1,    -- Valid values are 1-72 (37-72 only valid when BRAM_SIZE="36Kb")
      READ_WIDTH => 32)     -- Valid values are 1-72 (37-72 only valid when BRAM_SIZE="36Kb")
    port map (
      DO => mem_rbuf_DO(i),         -- Output read data port, width defined by READ_WIDTH parameter
      DI => mem_rbuf_DI(i),         -- Input write data port, width defined by WRITE_WIDTH parameter
      RDADDR => ipb_addr(9 downto 0), -- Input read address, width defined by read port depth
      RDCLK => sysclk,   -- 1-bit input read clock
      RDEN => '1',     -- 1-bit input read port enable
      REGCE => '0',   -- 1-bit input read output register enable
      RST => '0',       -- 1-bit input reset 
      WE => mem_rbuf_we,         -- Input write enable, width defined by write port depth
      WRADDR => mem_rbuf_wa, -- Input write address, width defined by write port depth
      WRCLK => ipb_clk,   -- 1-bit input write clock
      WREN => rbuf_wren      -- 1-bit input write port enable
   );
end generate;
process(ipb_addr)
begin
	if(ipb_addr(15 downto 13) = "100")then
		case ipb_addr(12 downto 10) is
			when "000" => ipb_rdata <= mem_rbuf_DO(0);
			when "001" => ipb_rdata <= mem_rbuf_DO(1);
			when "010" => ipb_rdata <= mem_rbuf_DO(2);
			when "011" => ipb_rdata <= mem_rbuf_DO(3);
			when "100" => ipb_rdata <= mem_rbuf_DO(4);
			when "101" => ipb_rdata <= mem_rbuf_DO(5);
			when "110" => ipb_rdata <= counter_rbuf_DO;
			when others =>
				if(ipb_addr(9 downto 4) = "000000")then
					case ipb_addr(3 downto 0) is
						when x"0" => ipb_rdata <= GO & zero(30 downto 3) & ipb_buf_NUM & ipb_read_all;
						when x"2" => ipb_rdata <= zero(31 downto N+1) & DIV;
						when x"3" => ipb_rdata <= zero(31 downto 15) & wbuf_ra;
						when x"4" => ipb_rdata <= zero(31 downto 15) & counter_rbuf_wa;
						when x"5" => ipb_rdata <= zero(31 downto 15) & mem_rbuf_wa;
						when others => ipb_rdata <= (others => '0');
					end case;
				elsif(ipb_addr(9 downto 4) = "000001")then
					ipb_rdata <= wbuf_DOA;
				else
					ipb_rdata <= (others => '0');
				end if;
		end case;
	else
		ipb_rdata <= (others => '0');
	end if;
end process;
end Behavioral;

